#!/usr/bin/perl
#=======================================================================
# * snmp_quagga_bgpd_mon.pl
# * v1.0a (2014.02.22)
# * by red_neon (red_neon@dcpp.ru)
# * Monitoring BGP session via Quagga daemon in Linux.
#=======================================================================
# CHANGES
#
#
# Monitoring BGP session via Quagga daemon in Linux v1.0a - 2014.02.22:
#
#	1. fixed: incorrect reading the string in command "show ip bgp summary" when the peer with State/PfxRcd: Idle (Admin)
#
#
# How to update previous version 1.0:
#
#	1. Replace old snmp_quagga_bgpd_mon.pl on new
#	2. Restart snmpd
#=======================================================================
# INSTALLATION:
#
#
# [REMOTE-SERVER SIDE]
#
#	[1] Quagga installed and "vtysh" properly configured using the nopassword directive.
#
#	[2] Copy quagga_bgpd_mon_v1.0a/remote_server/snmp_quagga_bgpd.pl to the remote server (for example /etc/srv_scripts/snmp_quagga_bgpd_mon.pl)
#	Then check path to "vtysh" in the script settings:
#	my $_vtysh = "/usr/bin/vtysh";
#
#	[3a] If SNMPD is running NOT from root user. And you don't want give access user "snmp" to "vtysh". You need to put into Cron next task:
#	$ nano /etc/crontab
#		*/5 * * * *  root /bin/sleep 290; /usr/bin/perl /etc/srv_scripts/snmp_quagga_bgpd_mon.pl cron 2>&1
#
#	Script will check Quagga via "vtysh" every 5 mins, then it will get data and save them into temp files in /tmp/ directory.
#	You can change directory and names of temp files in script-settings below (my $_file_.... )
#	To create temp files at first time, run the script in command line with parameter "cron":
#
#	$ /usr/bin/perl /etc/srv_scripts/snmp_quagga_bgpd_mon.pl cron
#
#	Ok. Go to pat.4)
#
#	[3b] If SNMPD is running from root user or snmp user have access to "vtysh". You need to change value of $_ROOT in this script below:
#	my $_ROOT = 1;
#
#	[4] Put into snmpd.conf string with pass_persist:
#	$ nano /etc/snmp/snmpd.conf
#		pass_persist .1.3.6.1.4.1.2021.190.1000.3 /usr/bin/perl /etc/srv_scripts/snmp_quagga_bgpd_mon.pl
#
#	[5] Restart snmpd:
#	$ /etc/init.d/snmpd restart
#
#	[6] Test via snmpwalk:
#	$ snmpwalk -v 2c -c public 127.0.0.1 .1.3.6.1.4.1.2021.190.1000.3
#	iso.3.6.1.4.1.2021.190.1000.3.1.0 = INTEGER: 1
#	iso.3.6.1.4.1.2021.190.1000.3.1.0.1 = INTEGER: 1
#	iso.3.6.1.4.1.2021.190.1000.3.1.1.1 = INTEGER: 1
#	iso.3.6.1.4.1.2021.190.1000.3.1.2.1 = STRING: "0.99.20.1"
#	iso.3.6.1.4.1.2021.190.1000.3.1.3.1 = STRING: "192.168.1.100"
#	iso.3.6.1.4.1.2021.190.1000.3.1.4.1 = STRING: "65000"
#	iso.3.6.1.4.1.2021.190.1000.3.1.5.1 = INTEGER: 17
#	iso.3.6.1.4.1.2021.190.1000.3.1.6.1 = INTEGER: 16
#	iso.3.6.1.4.1.2021.190.1000.3.1.7.1 = Gauge32: 5787558
#	iso.3.6.1.4.1.2021.190.1000.3.1.8.1 = Gauge32: 475307
#	iso.3.6.1.4.1.2021.190.1000.3.1.9.1 = Gauge32: 872419
#	iso.3.6.1.4.1.2021.190.1000.3.1.10.1 = Gauge32: 185567
#	iso.3.6.1.4.1.2021.190.1000.3.1.11.1 = Gauge32: 11023
#	iso.3.6.1.4.1.2021.190.1000.3.2.0 = INTEGER: 1
#	iso.3.6.1.4.1.2021.190.1000.3.2.1.1 = INTEGER: 1
#	iso.3.6.1.4.1.2021.190.1000.3.2.2.1 = INTEGER: 4
#	iso.3.6.1.4.1.2021.190.1000.3.2.3.1 = STRING: "[4] Neighbors"
#	...
#
#
# [CACTI-SERVER SIDE]
#
#	[7] Copy snmp_quagga_bgpd_neighbors_queries.xml and snmp_quagga_bgpd_router_queries.xml
#	from quagga_bgpd_mon_v1.0a/cacti_dir/resource/snmp_queries/
#	to: cacti/resource/snmp_queries/
#
#	[8] Import templates from quagga_bgpd_mon_v1.0a/cacti_web_templates/ in Cacti.
#=======================================================================
#	Sources:
#	snmp.pl		http://www.docum.org/stef.coene/qos/gui/
#	trafgraf.pl		by Sébastien Cramatte <scramatte@zensoluciones.com> October 2006
#	quagga-snmp-bgpd.in	v 1.3 by Oliver Hitz <oliver@net-track.ch> http://www.net-track.ch/opensource/quagga-snmp/
#=======================================================================
use strict;
#=======================================================================
# Settings:

# SNMPD is running from root user or snmp user have access to "vtysh" (0 = NO | 1 = YES)
my $_ROOT = 0;

# Path to "vtysh"
my $_vtysh = "/usr/bin/vtysh";

# Temp files for Cron-mode
my $_file_ipsum = "/tmp/snmp_quagga_bgpd_ipsum";
my $_file_rosum = "/tmp/snmp_quagga_bgpd_rosum";
my $_file_mem = "/tmp/snmp_quagga_bgpd_mem";
my $_file_ver = "/tmp/snmp_quagga_bgpd_ver";
my $_file_ipnei = "/tmp/snmp_quagga_bgpd_ipnei";

# OID. If you change it, you will need to change OIDs in xml-files in /cacti/resource/snmp_queries/
my $baseoid = '.1.3.6.1.4.1.2021.190.1000.3';
# How long first request will be cached
my $cache_secs = 4;

# Settings END
#=======================================================================
my $mibtime = 0;
my ($mib,$oid_in,$oid, $found, %cmds);
#=======================================================================
# Cron-mode
#=======================================================================
if (exists($ARGV[0])) {
	if ($ARGV[0] =~ /^cron$/) {
		get_data(1);
	} else {
		print("Bad param1!\n");
	}
	exit 0;
}
#=======================================================================
# Basic functions
#=======================================================================
# Switch on autoflush
$| = 1;

while (my $cmd = <STDIN>) {
	chomp $cmd;

	if ($cmd eq "PING") {
		print "PONG\n";
	} elsif ($cmd eq "get") {
		$oid_in = <STDIN>;
		$oid = get_oid($oid_in); # .1.3.6.1.4.1.2021.190.1000.3 -> 3
		$mib = create_mib(); 

		if (defined($mib->{$oid})) {
			# ANSWER:
			# current oid
			# type of value
			# value
			print "$baseoid.$oid\n";
			print $mib->{$oid}[0]."\n";
			print $mib->{$oid}[1]."\n";
		} else {
			print "NONE\n";
		}
	} elsif ($cmd eq "getnext") {
		$oid_in = <STDIN>;
		$oid = get_oid($oid_in);
		$mib = create_mib(); 
		$found = 0;

		my @s = sort {oidcmp($a, $b)} keys %{$mib};
		for (my $i = 0; $i < @s; $i++)
		{
			if (oidcmp($oid, $s[$i]) == -1)
			{
				# ANSWER:
				# next oid
				# type of value
				# value
				print "$baseoid.".$s[$i]."\n";
				print $mib->{$s[$i]}[0]."\n";
				print $mib->{$s[$i]}[1]."\n";
				$found = 1;
				last;
			}
		}

		if (!$found) {
			print "NONE\n";
		}    
	} else {
		# Unknown command
	}
}
exit 0;

sub get_oid
{
	my ($oid) = @_;
	chomp $oid;
	my $base = $baseoid;
	$base =~ s/\./\\./g;
	
	if ($oid !~ /^$base(\.|$)/) {
		# Requested oid doesn't match base oid
		return 0;
	}

	$oid =~ s/^$base\.?//;
	return $oid;
}

sub oidcmp {
	my ($x, $y) = @_;
	my @a = split /\./, $x;
	my @b = split /\./, $y;
	my $i = 0;

	while (1) {
		if ($i > $#a) {
			if ($i > $#b) {
				return 0;
			} else {
				return -1;
			}
		} elsif ($i > $#b) {
			return 1;
		}

		if ($a[$i] < $b[$i]) {
			return -1;
		} elsif ($a[$i] > $b[$i]) {
			return 1;
		}
		$i++;
	}
}

#=======================================================================
# Main functions
#=======================================================================
sub get_data
{
	my $_type = 0; # 0 = get data \ 1 = save data
	if (exists($_[0])) { $_type = $_[0] };
	my %_cmds = (
		"quagga" => 		{ "output" => [], "cmd" => [ $_file_ver,	"$_vtysh -c \"show version\" 2>&1|" ]},
		"neighbors" => 		{ "output" => [], "cmd" => [ $_file_ipnei,	"$_vtysh -c \"show ip bgp neighbors\" 2>&1|" ]},
		"neighbors_sum" => 	{ "output" => [], "cmd" => [ $_file_ipsum,	"$_vtysh -c \"show ip bgp summary\" 2>&1|" ]},
		"fib" => 		{ "output" => [], "cmd" => [ $_file_rosum,	"$_vtysh -c \"show ip route summary\" 2>&1|" ]},
		"rib" => 		{ "output" => [], "cmd" => [ $_file_mem,	"$_vtysh -c \"show memory\" 2>&1|" ]},
	);

	foreach my $key (keys %_cmds)
	{
		if (($_ROOT == 1) || ($_type == 1))
		{
			if (open(PROC, $_cmds{$key}{'cmd'}[1]))
			{
				@{$_cmds{$key}{'output'}} = <PROC>;
				close(PROC);
			}
		} else {
			if (open(PROC, '<', $_cmds{$key}{'cmd'}[0]))
			{
				@{$_cmds{$key}{'output'}} = <PROC>;
				close(PROC);
			}
		}
	}

	if ($_type == 1)
	{
		foreach my $key (keys %_cmds)
		{
			if (open(PROC, '>', $_cmds{$key}{'cmd'}[0]))
			{
				print(PROC @{$_cmds{$key}{'output'}});
				close(PROC);
			}
		}
	} else {
		%cmds = %_cmds;
	}
}

sub create_mib {
	if (time - $mibtime < $cache_secs) {return $mib;}

	my %snmp_res = (
		"1.0" => [ "Integer", 1 ],
		"1.0.1" => [ "Integer", 1 ],
		"1.1.1" => [ "Integer", 2 ],
		"1.2.1" => [ "String", "unknown" ],

		"1.3.1" => [ "String", "n/a" ],
		"1.4.1" => [ "String", "n/a" ],

		"1.5.1" => [ "Integer", 0 ],
		"1.6.1" => [ "Integer", 0 ],
		"1.7.1" => [ "Gauge32", 0 ],
		
		"1.8.1" => [ "Gauge32", 0 ],
		"1.9.1" => [ "Gauge32", 0 ],
		"1.10.1" => [ "Gauge32", 0 ],
		"1.11.1" => [ "Gauge32", 0 ],

		"2.0" => [ "Integer", 0 ],
		"3.0" => [ "Integer", 0 ],
				
	);

	######################################
	get_data(0);
	######################################
	if ( defined($cmds{'quagga'}{'output'}[0]) && ( $cmds{'quagga'}{'output'}[0] =~ /^Quagga\s+(\S+)/ ) )
	{
		$snmp_res{"1.1.1"}[1] = 1;
		$snmp_res{"1.2.1"}[1] = $1;
	}else{
		$mib = \%snmp_res;
		$mibtime = time;
		return $mib;
	}
	######################################
	foreach my $line (@{$cmds{'fib'}{'output'}})
	{
		if ( $line =~ /^(Totals)\s+\d+\s+(\d+)\s*$/ ){
			$snmp_res{'1.8.1'}[1] += $2;
		}
	}
	######################################
	foreach my $line (@{$cmds{'rib'}{'output'}})
	{
		if ( $line =~ /^BGP\s+aspath\s+:\s*(\d+)/ ){
			$snmp_res{'1.10.1'}[1] = $1;
		} elsif ( $line =~ /^BGP\s+node\s+:\s*(\d+)/ ) {
			$snmp_res{'1.9.1'}[1] = $1;
		} elsif ( $line =~ /^community\s+:\s*(\d+)/ ) {
			$snmp_res{'1.11.1'}[1] = $1;
		}
	}
	######################################
	my %_neighbors = ();
	my $_curpos = 0;
	foreach my $line (@{$cmds{'neighbors'}{'output'}})
	{
		if ( $line =~ /^BGP neighbor is (\S+), remote AS (\d+)/ )
		{
			$_neighbors{$1} = {"asn" => $2, "upd_sent" => 0, "upd_rcvd" => 0 };
			$_curpos = $1;
		} elsif ( $line =~ /^\s+Updates:\s+(\d+)\s+(\d+)/ ) {
			if (exists($_neighbors{$_curpos}))
			{
				$_neighbors{$_curpos}{"upd_sent"} = $1;
				$_neighbors{$_curpos}{"upd_rcvd"} = $2;
			}
		}
	}
	######################################
	my $_found = 0;
	foreach my $line (@{$cmds{'neighbors_sum'}{'output'}})
	{
		if ( $line =~ /BGP router identifier (\S+), local AS number (\d+)/ )
		{
			$snmp_res{'1.3.1'}[1] = $1;
			$snmp_res{'1.4.1'}[1] = $2;
		} elsif ( $line =~ /^Neighbor\s+V\s+AS\s+MsgRcvd\s+MsgSent\s+TblVer\s+InQ\s+OutQ\s+Up\/Down\s+State\/PfxRcd\s*$/ ) {
			$_found = 1;
		} elsif ($_found == 1) {
			my @n = split(/\s+/, $line);
			if (scalar(@n) >= 10)
			{
				$snmp_res{'1.5.1'}[1]++;
				my $_peer = $snmp_res{'1.5.1'}[1];
				my $_curlist = $snmp_res{"2.0"}[1];
				my $_peer_index = ($_peer - (8 * int($_peer / 8)));
				if ($_peer_index == 0){ $_peer_index = 8; }
				if ( ($snmp_res{'1.5.1'}[1] - (8 * $snmp_res{"2.0"}[1])) == 1 )
				{
					$snmp_res{"2.0"}[1]++;
					$_curlist = $snmp_res{"2.0"}[1];
					$snmp_res{"2.1.${_curlist}"} = [ "Integer", $_curlist ];
					$snmp_res{"2.2.${_curlist}"} = [ "Integer", 0 ];
					$snmp_res{"2.3.${_curlist}"} = [ "String", "[4] Neighbors" ];
					$snmp_res{"2.4.${_curlist}"} = [ "String", $snmp_res{'1.1.1'}[1] ];
					$snmp_res{"2.5.${_curlist}"} = [ "String", $snmp_res{'1.2.1'}[1] ];
					$snmp_res{"2.6.${_curlist}"} = [ "String", $snmp_res{'1.3.1'}[1] ];
					$snmp_res{"2.7.${_curlist}"} = [ "String", $snmp_res{'1.4.1'}[1] ];
				}
				if ($_peer_index > 4){
					$snmp_res{"2.3.${_curlist}"}[1] = "[8] Neighbors";
				}
				$snmp_res{"2.2.${_curlist}"}[1]++;
				########################################################
				my ($_desc, $_ip, $_asn, $_idx) = ("", $n[0], $n[2], $_peer);
				if (length($_ip) < 15)
				{
					my $_needlen = 15 - length($_ip);
					for (my $i = 1; $i <= $_needlen; $i++)
					{
						if ($i % 2 == 0) { $_ip = $_ip." " }
						else { $_ip = " ".$_ip }
					}
				}
				if (length($_asn) < 5)
				{
					my $_needlen = 5 - length($_asn);
					for (my $i = 1; $i <= $_needlen; $i++){
						$_asn = " ".$_asn
					}
				}
				if ($_idx == 9) { $_idx = " 9" }
				$_desc = "[$_ip][$_asn][$_idx]";
				########################################################
				$snmp_res{"2.10.${_peer_index}.${_curlist}"} = [ "Integer", $_peer ];
				$snmp_res{"2.11.${_peer_index}.${_curlist}"} = [ "String", $_desc ];
				$snmp_res{"2.12.${_peer_index}.${_curlist}"} = [ "Timeticks", 0 ];
				$snmp_res{"2.13.${_peer_index}.${_curlist}"} = [ "Gauge32", 0 ];
				$snmp_res{"2.14.${_peer_index}.${_curlist}"} = [ "Integer", -1 ];
				$snmp_res{"2.15.${_peer_index}.${_curlist}"} = [ "Counter32", 0 ];
				$snmp_res{"2.16.${_peer_index}.${_curlist}"} = [ "Counter32", 0 ];
				$snmp_res{"2.17.${_peer_index}.${_curlist}"} = [ "Integer", -1 ];

				$snmp_res{"3.0"}[1]++;
				$snmp_res{"3.1.${_peer}"} = [ "Integer", $_peer ];
				$snmp_res{"3.2.${_peer}"} = [ "String", $n[0] ];
				$snmp_res{"3.3.${_peer}"} = [ "Gauge32", $n[2] ];
				$snmp_res{"3.4.${_peer}"} = [ "Timeticks", 0 ];
				$snmp_res{"3.5.${_peer}"} = [ "Gauge32", 0 ];
				$snmp_res{"3.6.${_peer}"} = [ "Integer", -1 ];
				$snmp_res{"3.7.${_peer}"} = [ "Counter32", 0 ];
				$snmp_res{"3.8.${_peer}"} = [ "Counter32", 0 ];
				$snmp_res{"3.9.${_peer}"} = [ "Integer", -1 ];
				
				if ($n[9] =~ /^(\d+)$/ )
				{
					$snmp_res{'1.6.1'}[1]++;
					$snmp_res{'1.7.1'}[1] += $1;

					$snmp_res{"2.12.${_peer_index}.${_curlist}"}[1] = uptime($n[8]);
					$snmp_res{"2.13.${_peer_index}.${_curlist}"}[1] = $1;
					if ( ($snmp_res{"2.13.${_peer_index}.${_curlist}"}[1] <= $snmp_res{'1.8.1'}[1]) && ($snmp_res{'1.8.1'}[1] != 0) )
					{
						$snmp_res{"2.14.${_peer_index}.${_curlist}"}[1] =
							$snmp_res{"2.13.${_peer_index}.${_curlist}"}[1] * 100 / $snmp_res{'1.8.1'}[1];
					}

					$snmp_res{"3.4.${_peer}"}[1] = uptime($n[8]);
					$snmp_res{"3.5.${_peer}"}[1] = $1;
					if ( ($snmp_res{"3.5.${_peer}"}[1] <= $snmp_res{'1.8.1'}[1]) && ($snmp_res{'1.8.1'}[1] != 0) )
					{
						$snmp_res{"3.6.${_peer}"}[1] = $snmp_res{"3.5.${_peer}"}[1] * 100 / $snmp_res{'1.8.1'}[1];
					}

					if ( exists($_neighbors{$n[0]}) )
					{
						$snmp_res{"2.15.${_peer_index}.${_curlist}"}[1] = adjust_to_32bit($_neighbors{$n[0]}{"upd_sent"});
						$snmp_res{"2.16.${_peer_index}.${_curlist}"}[1] = adjust_to_32bit($_neighbors{$n[0]}{"upd_rcvd"});
						
						my $_pulse = $_neighbors{$n[0]}{"upd_sent"} + $_neighbors{$n[0]}{"upd_rcvd"};
						if ( $_pulse =~ /(\d{1,2})$/ ) { $_pulse = $1; } else { $_pulse = -1; }
						$snmp_res{"2.17.${_peer_index}.${_curlist}"}[1] = $_pulse;
						
						$snmp_res{"3.7.${_peer}"}[1] = adjust_to_32bit($_neighbors{$n[0]}{"upd_sent"});
						$snmp_res{"3.8.${_peer}"}[1] = adjust_to_32bit($_neighbors{$n[0]}{"upd_rcvd"});
						$snmp_res{"3.9.${_peer}"}[1] = $_pulse;
					}
				}
			} else {
				$_found = 0;
			}
		}
	}
	######################################
	my $_curlist = $snmp_res{"2.0"}[1];
	my $_tot_inlist = $snmp_res{"2.2.".$_curlist}[1];
	my @_graphtype = $snmp_res{"2.3.".$_curlist}[1] =~ m/^\[(\d+)\]/;
	if ( ($_graphtype[0] =~ m/(4|8)/) && ($_tot_inlist < $_graphtype[0]) ) {
		my $_need = $_graphtype[0] - $_tot_inlist;
		my $_desc = "---------------------------";
		if ($_curlist > 1){
			$_desc = "----------------------------";
		}
		for (my $i=($_tot_inlist+1); $i<=$_graphtype[0]; $i++){
			$snmp_res{"2.11.".$i.".".$_curlist} = [ "String", $_desc ];
		}
	}
	######################################
	$mib = \%snmp_res;
	$mibtime = time;
	return $mib;
} 

#=======================================================================
# Secondary functions
#=======================================================================

sub uptime
{
	my ($t) = @_;
	if ($t =~ /^(\d+):(\d+):(\d+)$/) {
		return 100*($3+60*($2+$1*60));
	} elsif ($t =~ /^(\d+)d(\d+)h(\d+)m$/) {
		return 100*60*($3+60*($2+24*$1));
	} elsif ($t =~ /^(\d+)w(\d+)d(\d+)h$/) {
		return 100*60*60*($3+24*$2+7*24*$1);
	}
	return 0;
}

sub adjust_to_32bit
{
	my ($val) = @_;
	if ($val > 4294967295)
	{
		$val = $val % 4294967295;
	}
	return $val;
}

#=======================================================================
